# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Renders policy source files into actual Access Control Lists."""

import copy
import multiprocessing.context
import multiprocessing.managers
import multiprocessing.pool
import pathlib
import sys
import typing
from collections.abc import Iterator

from absl import app, flags, logging

from aerleon.lib import aclgenerator, naming, pcap, plugin_supervisor, policy, yaml
from aerleon.utils import config

FLAGS = flags.FLAGS
WriteList = typing.MutableSequence[tuple[pathlib.Path, str]]


def SetupFlags():
    """Read in configuration from CLI flags."""
    flags.DEFINE_string(
        'base_directory',
        None,
        'The base directory to search recursively for policy files.\n'
        'Relative policy imports are resolved against this directory.\n'
        'If --policy_file is used, aclgen will not search this directory.\n'
        'Default: \'%s\'' % config.defaults['base_directory'],
    )
    flags.DEFINE_string(
        'definitions_directory',
        None,
        'Directory containing network and service definition files.\n'
        'Default: \'%s\'' % config.defaults['definitions_directory'],
    )
    flags.DEFINE_string('policy_file', None, 'Individual policy file to generate.')
    flags.DEFINE_string(
        'output_directory',
        None,
        'Directory to output the rendered acls.\nDefault: \'%s\''
        % config.defaults['output_directory'],
    )
    flags.DEFINE_boolean(
        'optimize',
        None,
        f"Turn on optimization.\nDefault: '{config.defaults['optimize']}'",
        short_name='o',
    )
    flags.DEFINE_boolean(
        'recursive',
        None,
        'UNUSED. '
        'Recursive policy file search is always enabled except when using the --policy_file flag.',
    )
    flags.DEFINE_boolean(
        'debug',
        None,
        f"Display detailed messages.\nDefault: '{str(config.defaults['debug']).lower()}'",
    )
    flags.DEFINE_boolean('verbose', None, 'UNUSED. Use --debug instead.')
    flags.DEFINE_list(
        'ignore_directories',
        None,
        'Don\'t descend into directories that look like this string.\nDefault: \'%s\''
        % ','.join(config.defaults['ignore_directories']),
    )
    flags.DEFINE_integer(
        'max_renderers',
        None,
        'Max number of rendering processes to use.\nDefault: \'%s\''
        % config.defaults['max_renderers'],
    )
    flags.DEFINE_boolean(
        'shade_check',
        None,
        'Raise an error when a term is completely shaded by a prior term.\nDefault: \'%s\''
        % str(config.defaults['shade_check']).lower(),
    )
    flags.DEFINE_integer(
        'exp_info',
        None,
        'Print a message when a term is set to expire in that many weeks.\nDefault: \'%s\''
        % str(config.defaults['exp_info']),
    )
    flags.DEFINE_multi_string('config_file', None, 'A YAML file with configuration options')


class Error(Exception):
    """Base Error class."""


class P4WriteFileError(Error):
    """Error when there are issues p4 editing the destination."""


class ACLGeneratorError(Error):
    """Raised when an ACL generator has errors."""


class ACLParserError(Error):
    """Raised when the ACL parser fails."""


def RenderFile(
    base_directory: str,
    input_file: pathlib.Path,
    output_directory: pathlib.Path,
    definitions: naming.Naming,
    exp_info: int,
    optimize: bool,
    shade_check: bool,
    write_files: WriteList,
):
    """Render a single file.

    Args:
      base_directory: The base directory to look for acls.
      input_file: the name of the input policy file.
      output_directory: the directory in which we place the rendered file.
      definitions: the definitions from naming.Naming().
      exp_info: print a info message when a term is set to expire in that many
        weeks.
      optimize: a boolean indicating if we should turn on optimization or not.
      shade_check: should we raise an error if a term is completely shaded
      write_files: a list of file tuples, (output_file, acl_text), to write
    """
    output_relative = input_file.relative_to(base_directory).parent.parent
    output_directory = output_directory / output_relative

    logging.debug('rendering file: %s into %s', input_file, output_directory)

    pol = None

    try:
        with open(input_file) as f:
            conf = f.read()
            logging.debug('opened and read %s', input_file)
    except OSError as e:
        logging.warning('bad file: \n%s', e)
        raise

    try:
        # PolicySource[extension].ParsePolicy(conf)
        if pathlib.Path(input_file).suffix == '.yaml' or pathlib.Path(input_file).suffix == '.yml':
            pol = yaml.ParsePolicy(
                conf,
                filename=input_file,
                base_dir=base_directory,
                definitions=definitions,
                optimize=optimize,
                shade_check=shade_check,
            )
        else:
            pol = policy.ParsePolicy(
                conf,
                definitions,
                optimize=optimize,
                base_dir=base_directory,
                shade_check=shade_check,
            )
    except policy.ShadingError as e:
        logging.warning('shading errors for %s:\n%s', input_file, e)
        return
    except (policy.Error, naming.Error) as e:
        raise ACLParserError(
            'Error parsing policy file %s:\n%s%s'
            % (input_file, sys.exc_info()[0], sys.exc_info()[1])
        ) from e

    platforms = set()
    for header in pol.headers:
        platforms.update(header.platforms)

    acl_obj: aclgenerator.ACLGenerator
    plugin_supervisor.PluginSupervisor.Start()

    for target in platforms:
        generator = plugin_supervisor.PluginSupervisor.generators.get(target)
        if not generator:
            logging.warning(f"No generator found for target \"{target}\", skipping target.")
            continue

        try:
            # special handling for pcap
            if target == 'pcap':
                assert issubclass(generator, pcap.PcapFilter)
                acl_obj = generator(copy.deepcopy(pol), exp_info)
                RenderACL(
                    str(acl_obj),
                    f"-accept{acl_obj.SUFFIX}",
                    output_directory,
                    input_file,
                    write_files,
                )
                acl_obj = generator(copy.deepcopy(pol), exp_info, invert=True)
                RenderACL(
                    str(acl_obj),
                    f"-deny{acl_obj.SUFFIX}",
                    output_directory,
                    input_file,
                    write_files,
                )
            else:
                acl_obj = generator(copy.deepcopy(pol), exp_info)
                RenderACL(str(acl_obj), acl_obj.SUFFIX, output_directory, input_file, write_files)

        except aclgenerator.Error as e:
            raise ACLGeneratorError(f'Error generating target ACL for {input_file}:\n{e}') from e


def RenderACL(
    acl_text: str,
    acl_suffix: str,
    output_directory: pathlib.Path,
    input_file: pathlib.Path,
    write_files: typing.MutableSequence[tuple[pathlib.Path, str]],
    binary: bool = False,
):
    """Write the ACL string out to file if appropriate.

    Args:
      acl_text: Rendered output of an ACL Generator.
      acl_suffix: File suffix to append to output filename.
      output_directory: The directory to write the output file.
      input_file: The name of the policy file that was used to render ACL.
      write_files: A list of file tuples, (output_file, acl_text), to write.
      binary: Boolean if the rendered ACL is in binary format.
    """
    input_filename = input_file.with_suffix(acl_suffix).name
    output_file = output_directory / input_filename

    if FilesUpdated(output_file, acl_text, binary):
        logging.info('file changed: %s', output_file)
        write_files.append((output_file, acl_text))
    else:
        logging.debug('file not changed: %s', output_file)


def FilesUpdated(file_name: pathlib.Path, new_text: str, binary: bool) -> bool:
    """Diff the rendered acl with what's already on disk.

    Args:
      file_name: Name of file on disk to check against.
      new_text: Text of newly generated ACL.
      binary: True if file is a binary format.

    Returns:
      Boolean if config does not equal new text.
    """
    if binary:
        readmode = 'rb'
    else:
        readmode = 'r'
    try:
        with open(file_name, readmode) as f:
            conf: str = str(f.read())
    except OSError:
        return True
    if not binary:
        p4_id = '$I d:'.replace(' ', '')
        p4_date = '$Da te:'.replace(' ', '')
        p4_revision = '$Rev ision:'.replace(' ', '')

        def P4Tags(text: str) -> bool:
            return not (p4_id in text or p4_date in text or p4_revision in text)

        filtered_conf = filter(P4Tags, conf.split('\n'))
        filtered_text = filter(P4Tags, new_text.split('\n'))
        return list(filtered_conf) != list(filtered_text)
    return conf != new_text


def DescendDirectory(input_dirname: str, ignore_directories: list[str]) -> list[pathlib.Path]:
    """Descend from input_dirname looking for policy files to render.

    Args:
      input_dirname: the base directory.
      ignore_directories: directories to ignore while traversing.

    Returns:
      a list of input file paths
    """
    input_dir = pathlib.Path(input_dirname)

    policy_files: list[pathlib.Path] = []
    policy_directories: Iterator[pathlib.Path] = filter(
        lambda path: path.is_dir(), input_dir.glob('**/pol')
    )
    for ignored_directory in ignore_directories:

        def Filtering(path, ignored=ignored_directory):
            return not path.match(f'{ignored}/**/pol') and not path.match(f'{ignored}/pol')

        policy_directories = filter(Filtering, policy_directories)

    for directory in policy_directories:
        # Build glob strings from PolicySources.keys()
        # Or just match by extension
        directory_policies = (
            list(directory.glob('*.pol'))
            + list(directory.glob('*.yaml'))
            + list(directory.glob('*.yml'))
        )
        depth = len(directory.parents) - 1
        logging.warning(
            f"{'-' * (2 * depth)}> {directory} ({len(directory_policies)} pol files found)"
        )
        policy_files.extend(filter(lambda path: path.is_file(), directory_policies))

    return policy_files


def WriteFiles(write_files: WriteList):
    """Writes files to disk.

    Args:
      write_files: List of file names and strings.
    """
    if write_files:
        logging.info('writing %d files to disk...', len(write_files))
    else:
        logging.info('no files changed, not writing to disk')
    for output_file, file_contents in write_files:
        _WriteFile(output_file, file_contents)


def _WriteFile(output_file: pathlib.Path, file_contents: str):
    """Inner file writing function.

    Args:
      output_file: Path to write to
      file_contents: Data to write
    """
    try:
        parent_path = pathlib.Path(output_file).parent
        if not parent_path.is_dir():
            parent_path.mkdir(parents=True, exist_ok=True)
        with open(output_file, 'w') as output:
            logging.info('writing file: %s', output_file)
            output.write(file_contents)
    except OSError:
        logging.warning('error while writing file: %s', output_file)
        raise


def Run(
    base_directory: str,
    definitions_directory: str,
    policy_file: str,
    output_directory: str,
    exp_info: int,
    max_renderers: int,
    ignore_directories: list[str],
    optimize: bool,
    shade_check: bool,
    context: multiprocessing.context.BaseContext,
):
    """Generate ACLs.

    Args:
      base_directory: directory containing policy files.
      definitions_directory: directory containing NETWORK and SERVICES definition
        files.
      policy_file: path to a single policy file to render.
      output_directory: directory in which rendered files are placed.
      exp_info: print a info message when a term is set to expire in that many
        weeks.
      max_renderers: the number of renderers to run in parallel.
      ignore_directories: directories to ignore when searching for policy files.
      optimize: a boolean indicating if we should turn on optimization or not.
      shade_check: should we raise an error if a term is completely shaded.
      context: multiprocessing context
    """
    definitions = None
    try:
        definitions = naming.Naming(definitions_directory)
    except naming.NoDefinitionsError:
        err_msg = f'bad definitions directory: {definitions_directory}'
        logging.fatal(err_msg)
        return  # static type analyzer can't detect that logging.fatal exits program

    # thead-safe list for storing files to write
    manager: multiprocessing.managers.SyncManager = context.Manager()
    write_files: WriteList = manager.list()

    with_errors = False
    logging.info('finding policies...')
    if max_renderers == 1 or policy_file:
        if policy_file:
            policies = [pathlib.Path(policy_file)]
        else:
            policies = DescendDirectory(base_directory, ignore_directories)
        try:
            for pol in policies:
                RenderFile(
                    base_directory,
                    pol,
                    pathlib.Path(output_directory),
                    definitions,
                    exp_info,
                    optimize,
                    shade_check,
                    write_files,
                )
        except (ACLParserError, ACLGeneratorError) as e:
            with_errors = True
            logging.warning('\n\nerror encountered in rendering process:\n%s\n\n', e)
    else:
        # render all files in parallel
        policies = DescendDirectory(base_directory, ignore_directories)
        pool = context.Pool(processes=max_renderers)
        results: list[multiprocessing.pool.AsyncResult] = []
        for pol in policies:
            results.append(
                pool.apply_async(
                    RenderFile,
                    args=(
                        base_directory,
                        pol,
                        output_directory,
                        definitions,
                        exp_info,
                        optimize,
                        shade_check,
                        write_files,
                    ),
                )
            )
        pool.close()
        pool.join()

        for result in results:
            try:
                result.get()
            except (ACLParserError, ACLGeneratorError) as e:
                with_errors = True
                logging.warning('\n\nerror encountered in rendering process:\n%s\n\n', e)

    # actually write files to disk
    WriteFiles(write_files)

    if with_errors:
        logging.warning('done, with errors.')
        sys.exit(1)
    else:
        logging.info('done.')


def main(argv):
    del argv  # Unused.

    absl_flags = {
        flag: getattr(FLAGS, flag) for flag in config.defaults.keys() if getattr(FLAGS, flag, None)
    }
    try:
        configs = config.load_config(config_file=FLAGS.config_file)
        configs.update(absl_flags)
    except config.ConfigFileError as e:
        exit(f"Error: {e}")

    if configs['verbose']:
        logging.set_verbosity(logging.INFO)
    if configs['debug']:
        logging.set_verbosity(logging.DEBUG)
    logging.debug(
        'binary: %s\noptimize: %d\nbase_directory: %s\n'
        'policy_file: %s\nrendered_acl_directory: %s',
        str(sys.argv[0]),
        int(configs['optimize']),
        str(configs['base_directory']),
        str(configs['policy_file']),
        str(configs['output_directory']),
    )
    logging.debug('aerleon configurations: %s', configs)

    context = multiprocessing.get_context()
    try:
        Run(
            configs['base_directory'],
            configs['definitions_directory'],
            configs['policy_file'],
            configs['output_directory'],
            configs['exp_info'],
            configs['max_renderers'],
            configs['ignore_directories'],
            configs['optimize'],
            configs['shade_check'],
            context,
        )
    except Exception as e:
        logging.error(f"Unhandled exception: {e}", exc_info=True)
        sys.exit(1)


def EntryPoint():
    """Read in flags and call main()."""
    SetupFlags()
    app.run(main)


if __name__ == '__main__':
    EntryPoint()



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A subclass of the ipaddress library that includes comments for ipaddress."""

from __future__ import annotations

import collections
import ipaddress
import itertools
from typing import Union

import aerleon.utils.iputils as iputils


def IP(
    ip: ipaddress.IPv4Network | ipaddress.IPv6Network | str,
    comment: str = '',
    token: str = '',
    strict: bool = True,
) -> IPv4 | IPv6:
    """Take an ip string and return an object of the correct type.

    Args:
      ip: the ip address.
      comment: option comment field
      token: option token name where this address was extracted from
      strict: If strict should be used in ipaddress object.

    Returns:
      ipaddress.IPv4 or ipaddress.IPv6 object or raises ValueError.

    Raises:
      ValueError: if the string passed isn't either a v4 or a v6 address.
    """
    if isinstance(ip, ipaddress._BaseNetwork):  # pylint disable=protected-access
        imprecise_ip = ip
    else:
        imprecise_ip = ipaddress.ip_network(ip, strict=strict)
    if imprecise_ip.version == 4:
        return IPv4(ip, comment, token, strict=strict)
    elif imprecise_ip.version == 6:
        return IPv6(ip, comment, token, strict=strict)
    raise ValueError(f'Provided IP string "{ip}" is not a valid v4 or v6 address')


# TODO(robankeny) remove once at 3.7
@staticmethod
def _is_subnet_of(a: IPv4 | IPv6, b: IPv4 | IPv6) -> bool:  # pylint: disable=invalid-name
    try:
        # Always false if one is v4 and the other is v6.
        if a.version != b.version:
            raise TypeError(f'{a} and {b} are not of the same version')
        return (
            b.network_address <= a.network_address and b.broadcast_address >= a.broadcast_address
        )
    except AttributeError:
        raise TypeError(f'Unable to test subnet containment between {a} and {b}')


class IPv4(ipaddress.IPv4Network):
    """This subclass allows us to keep text comments related to each object."""

    def __init__(
        self,
        ip_string: ipaddress.IPv4Network | tuple[int, int] | str | IPv4,
        comment: str | IPv4 = '',
        token: str = '',
        strict: bool = True,
    ) -> None:
        self.text = comment
        self.token = token
        self.parent_token = token

        # Using a tuple of IP integer/prefixlength is significantly faster than
        # using the BaseNetwork object for recreating the IP network
        if isinstance(ip_string, ipaddress._BaseNetwork):  # pylint disable=protected-access
            ip = (
                ip_string.network_address._ip,
                ip_string.prefixlen,
            )  # pylint disable=protected-access # pytype: disable=attribute-error
        else:
            ip = ip_string
        super().__init__(ip, strict)

    def subnet_of(self, other: IPv4) -> bool:
        """Return True if this network is a subnet of other."""
        if self.version != other.version:
            return False
        return self._is_subnet_of(self, other)

    def supernet_of(self, other: IPv4 | IPv6) -> bool:
        """Return True if this network is a supernet of other."""
        if self.version != other.version:
            return False
        return self._is_subnet_of(other, self)

    def __deepcopy__(self, memo: dict):
        result = self.__class__(self)
        result.text = self.text
        result.token = self.token
        result.parent_token = self.parent_token
        return result

    def AddComment(self, comment: str = '') -> None:
        """Append comment to self.text, comma separated.

        Don't add the comment if it's the same as self.text.

        Args:
          comment: comment to be added.
        """
        if self.text:
            if comment and comment not in self.text:
                self.text += f", {comment}"
        else:
            self.text = comment

    def supernet(self, prefixlen_diff: int = 1) -> IPv4:
        """Override ipaddress.IPv4 supernet so we can maintain comments.

        See ipaddress.IPv4.Supernet for complete documentation.

        Args:
          prefixlen_diff: Prefix length difference.

        Returns:
          An IPv4 object

        Raises:
          PrefixlenDiffInvalidError: Raised when prefixlen - prefixlen_diff results
            in a negative number.
        """
        if self.prefixlen == 0:
            return self
        if self.prefixlen - prefixlen_diff < 0:
            raise PrefixlenDiffInvalidError(
                'current prefixlen is %d, cannot have a prefixlen_diff of %d'
                % (self.prefixlen, prefixlen_diff)
            )
        ret_addr = IPv4(
            ipaddress.IPv4Network.supernet(self, prefixlen_diff),
            comment=self.text,
            token=self.token,
        )
        return ret_addr

    # Backwards compatibility name from v1.
    Supernet = supernet


class IPv6(ipaddress.IPv6Network):
    """This subclass allows us to keep text comments related to each object."""

    def __init__(
        self,
        ip_string: str | ipaddress.IPv6Network | tuple[int, int] | IPv6,
        comment: str = '',
        token: str = '',
        strict: bool = True,
    ) -> None:
        self.text = comment
        self.token = token
        self.parent_token = token

        # Using a tuple of IP integer/prefixlength is significantly faster than
        # using the BaseNetwork object for recreating the IP network
        if isinstance(ip_string, ipaddress._BaseNetwork):  # pylint disable=protected-access
            ip = (
                ip_string.network_address._ip,
                ip_string.prefixlen,
            )  # pylint disable=protected-access # pytype: disable=attribute-error
        else:
            ip = ip_string
        super().__init__(ip, strict)

    def subnet_of(self, other: IPv6) -> bool:
        """Return True if this network is a subnet of other."""
        if self.version != other.version:
            return False
        return self._is_subnet_of(self, other)

    def supernet_of(self, other: IPv6) -> bool:
        """Return True if this network is a supernet of other."""
        if self.version != other.version:
            return False
        return self._is_subnet_of(other, self)

    def __deepcopy__(self, memo: dict):
        result = self.__class__(self)
        result.text = self.text
        result.token = self.token
        result.parent_token = self.parent_token
        return result

    def supernet(self, prefixlen_diff: int = 1) -> IPv6:
        """Override ipaddress.IPv6Network supernet so we can maintain comments.

        See ipaddress.IPv6Network.Supernet for complete documentation.
        Args:
          prefixlen_diff: Prefix length difference.

        Returns:
          An IPv4 object

        Raises:
          PrefixlenDiffInvalidError: Raised when prefixlen - prefixlen_diff results
            in a negative number.
        """
        if self.prefixlen == 0:
            return self
        if self.prefixlen - prefixlen_diff < 0:
            raise PrefixlenDiffInvalidError(
                'current prefixlen is %d, cannot have a prefixlen_diff of %d'
                % (self.prefixlen, prefixlen_diff)
            )
        ret_addr = IPv6(
            ipaddress.IPv6Network.supernet(self, prefixlen_diff),
            comment=self.text,
            token=self.token,
        )
        return ret_addr

    # Backwards compatibility name from v1.
    Supernet = supernet

    def AddComment(self, comment: str = '') -> None:
        """Append comment to self.text, comma separated.

        Don't add the comment if it's the same as self.text.

        Args:
          comment: comment to be added.
        """
        if self.text:
            if comment and comment not in self.text:
                self.text += f", {comment}"
        else:
            self.text = comment


IPType = Union[IPv4, IPv6]


def _InNetList(adders: list[ipaddress._BaseNetwork], ip: ipaddress._BaseNetwork) -> bool:
    """Returns True if ip is contained in adders."""
    for addr in adders:
        if ip.subnet_of(addr):
            return True
    return False


def IsSuperNet(
    supernets: list[ipaddress._BaseNetwork], subnets: list[ipaddress._BaseNetwork]
) -> bool:
    """Returns True if subnets are fully consumed by supernets."""
    for net in subnets:
        if not _InNetList(supernets, net):
            return False
    return True


def CollapseAddrListPreserveTokens(addresses: list[IPv4 | IPv6]) -> list[IPv4 | IPv6]:
    """Collapse an array of IPs only when their tokens are the same.

    Args:
       addresses: list of ipaddress.IPNetwork objects.

    Returns:
      list of ipaddress.IPNetwork objects.
    """
    ret_array = []
    for grp in itertools.groupby(
        sorted(addresses, key=lambda x: x.parent_token), lambda x: x.parent_token
    ):
        ret_array.append(CollapseAddrList(list(grp[1])))
    dedup_array = []
    i = 0
    while len(ret_array) > i:
        ip = ret_array.pop(0)
        k = 0
        to_add = True
        while k < len(dedup_array):
            if IsSuperNet(dedup_array[k], ip):
                to_add = False
                break
            elif IsSuperNet(ip, dedup_array[k]):
                del dedup_array[k]
            k += 1
        if to_add:
            dedup_array.append(ip)
    return [i for sublist in dedup_array for i in sublist]


def _SafeToMerge(
    address: IPv4 | IPv6,
    merge_target: IPv4 | IPv6,
    check_addresses: (
        dict[ipaddress.IPv4Address, list[IPv4]] | dict[ipaddress.IPv6Address, list[IPv6]]
    ),
) -> bool:
    """Determine if it's safe to merge address into merge target.

    Checks given address against merge target and a list of check_addresses
    if it's OK to roll address into merge target such that it not less specific
    than any of the check_addresses. See description of why ir is important
    within public function CollapseAddrList.

    Args:
      address: Address that is being merged.
      merge_target: Merge candidate address.
      check_addresses: A dict networks_address->addrs to compare specificity with.

    Returns:
      True if safe to merge, False otherwise.
    """
    for check_address in check_addresses.get(address.network_address, []):
        if merge_target.netmask <= check_address.netmask < address.netmask:
            return False
    return True


def _CollapseAddrListInternal(
    addresses: list[IPv4 | IPv6],
    complements_by_network: (
        dict[ipaddress.IPv4Address, list[IPv4]] | dict[ipaddress.IPv6Address, list[IPv6]]
    ),
) -> list[IPv4 | IPv6]:
    """Collapses consecutive netblocks until reaching a fixed point.

     Example:

     ip1 = ipaddress.IPv4Network('1.1.0.0/24')
     ip2 = ipaddress.IPv4Network('1.1.1.0/24')
     ip3 = ipaddress.IPv4Network('1.1.2.0/24')
     ip4 = ipaddress.IPv4Network('1.1.3.0/24')
     ip5 = ipaddress.IPv4Network('1.1.4.0/24')
     ip6 = ipaddress.IPv4Network('1.1.0.1/22')

     _CollapseAddrListInternal([ip1, ip2, ip3, ip4, ip5, ip6]) ->
     [IPv4Network('1.1.0.0/22'), IPv4Network('1.1.4.0/24')]

     Note, this shouldn't be called directly, but is called via
     CollapseAddrList([])

    Args:
      addresses: List of IPv4 or IPv6 objects
      complements_by_network: Dict of IPv4 or IPv6 objects indexed by
        network_address, that if present will be considered to avoid harmful
        optimizations.

    Returns:
      List of IPv4 or IPv6 objects (depending on what we were passed)
    """
    ret_array = []
    for addr in addresses:
        addr_is_fresh = True
        while addr_is_fresh:
            addr_is_fresh = False
            if not ret_array:
                ret_array.append(addr)
                continue

            prev_addr = ret_array[-1]
            if not _SafeToMerge(addr, prev_addr, complements_by_network):
                ret_array.append(addr)
            elif prev_addr.supernet_of(addr):
                # Preserve addr's comment, then subsume it.
                prev_addr.AddComment(addr.text)
            elif (
                prev_addr.version == addr.version
                and prev_addr.prefixlen == addr.prefixlen
                and
                # It's faster to compare integers than IP objects
                prev_addr.broadcast_address._ip + 1
                == addr.network_address._ip  # pylint disable=protected-access
                and
                # Generating Supernet is relatively intensive compared to doing bit
                # operations
                (prev_addr.netmask._ip << 1) & prev_addr.network_address._ip
                == prev_addr.network_address._ip  # pylint disable=protected-access
            ):  # pylint disable=protected-access
                # Preserve addr's comment, then merge with it.
                prev_addr.AddComment(addr.text)
                addr = ret_array.pop().Supernet()
                addr_is_fresh = True
            else:
                ret_array.append(addr)

    return ret_array


def CollapseAddrList(
    addresses: list[IPv4 | IPv6],
    complement_addresses: list[IPv4] | list[IPv6] | None = None,
) -> list[IPv4 | IPv6]:
    """Collapse an array of IP objects.

    Example:  CollapseAddrList(
      [IPv4('1.1.0.0/24'), IPv4('1.1.1.0/24')]) -> [IPv4('1.1.0.0/23')]
      Note: this works just as well with IPv6 addresses too.

    On platforms that support exclude semantics with most specific match,
    this method should _always_ be called with complement addresses supplied.
    Not doing so can lead to *reversal* of intent. Consider this case:

      destination-address:: 10.0.0.0/8, 10.0.0.0/10
      destination-exclude:: 10.0.0.0/9

    Without optimization, 10.0.0.1 will _match_. With optimization, most specific
    prefix will _not_ match, reversing the intent. Supplying complement_addresses
    allows this method to consider those implications.

    Args:
       addresses: list of ipaddress.IPNetwork objects
       complement_addresses: list of ipaddress.IPNetwork objects that, if present,
        will be considered to avoid harmful optimizations.

    Returns:
      list of ipaddress.IPNetwork objects
    """
    complements_dict = collections.defaultdict(list)
    address_set = {a.network_address for a in addresses}
    for ca in complement_addresses or []:
        if ca.network_address in address_set:
            complements_dict[ca.network_address].append(ca)
    return _CollapseAddrListInternal(
        sorted(addresses, key=ipaddress.get_mixed_type_key), complements_dict
    )


def SortAddrList(addresses: list[IPv6 | IPv4]) -> list[IPv6 | IPv4]:
    """Return a sorted list of nacaddr objects."""
    return sorted(addresses, key=ipaddress.get_mixed_type_key)


def RemoveAddressFromList(superset: list[IPv4 | IPv6], exclude: IPv4 | IPv6) -> list[IPv6 | IPv4]:
    """Remove a single address from a list of addresses.

    Args:
      superset: a List of nacaddr IPv4 or IPv6 addresses
      exclude: a single nacaddr IPv4 or IPv6 address

    Returns:
      a List of nacaddr IPv4 or IPv6 addresses
    """
    ret_array = []
    for addr in superset:
        if exclude == addr or addr.subnet_of(exclude):
            pass
        elif exclude.version == addr.version and exclude.subnet_of(addr):
            # this could be optimized except that one group uses this
            # code with ipaddrs (instead of nacaddrs).
            ret_array.extend(IP(x) for x in iputils.exclude_address(addr, exclude))
        else:
            ret_array.append(addr)
    return SortAddrList(ret_array)


def AddressListExclude(
    superset: list[IPv4 | IPv6],
    excludes: list[IPv4 | IPv6],
    collapse_addrs: bool = True,
) -> list[IPv4 | IPv6]:
    """Remove a list of addresses from another list of addresses.

    Args:
      superset: a List of nacaddr IPv4 or IPv6 addresses
      excludes: a List nacaddr IPv4 or IPv6 addresses
      collapse_addrs: whether or not to collapse contiguous CIDRs togethe

    Returns:
      a List of nacaddr IPv4 or IPv6 addresses
    """
    if collapse_addrs:
        superset = CollapseAddrList(superset)[::-1]
        excludes = CollapseAddrList(excludes)[::-1]
    else:
        superset = sorted(superset, reverse=True)
        excludes = sorted(excludes, reverse=True)

    ret_array = []
    while superset and excludes:
        if superset[-1].overlaps(excludes[-1]):
            ip = superset.pop()
            superset.extend(reversed(RemoveAddressFromList([ip], excludes[-1])))
        elif (
            superset[-1]._get_networks_key() < excludes[-1]._get_networks_key()
        ):  # pylint: disable=protected-access
            ret_array.append(superset.pop())
        else:
            excludes.pop()
    if collapse_addrs:
        return CollapseAddrList(ret_array + superset)
    else:
        return sorted(set(ret_array + superset))


ExcludeAddrs = AddressListExclude


class PrefixlenDiffInvalidError(ipaddress.NetmaskValueError):
    """Holdover from ipaddr v1."""


if __name__ == '__main__':
    pass



from collections import defaultdict
from collections.abc import MutableMapping
from dataclasses import dataclass
from typing import Any

from absl import logging

from aerleon.lib import aclgenerator, nacaddr, policy
from aerleon.lib.proxmox import BooleanKeywordOption
from aerleon.utils.options import (
    AbstractOption,
    ArbitraryValueOption,
    ProcessOptions,
    ValueOption,
)

FORTIGATE_SERVICES_ALL = 'ALL'
FORTIGATE_ADDRESS_ALL = 'all'
FORTIGATE_ADDRESS_NONE = 'none'
FORTIGATE_SCHEDULE_ALWAYS = '"always"'
SUPPORTED_ACTIONS = {'accept', 'deny'}
FORTIGATE_COMMENT_LIMIT = 1023
FORTIGATE_LOG_TRAFFIC_VALUES = {'log_traffic_mode_all', 'log_traffic_mode_disable'}
FORTIGATE_LOG_TRAFFIC_START_VALUES = {'log_traffic_start_session'}


class FortigateDefaultDictionary(defaultdict):
    def __init__(self, object_constructor, key_attribute_name):
        """
        Initializes the FortigateDefaultDictionary.

        Args:
            object_constructor: The class (or function) to call to create a new object.
                                Its __init__ or call signature must accept a keyword
                                argument named by key_attribute_name.
            key_attribute_name: The string name of the attribute in the object
                                that should be populated with the dictionary key.
        """
        super().__init__(None)  # Initialize defaultdict without a default_factory
        # We handle creation in __missing__
        if not callable(object_constructor):
            raise TypeError("object_constructor must be callable")
        self.object_constructor = object_constructor
        self.key_attribute_name = key_attribute_name

    def __missing__(self, key):
        """
        Called when a key is not found. Creates the object using the key.
        """
        kwargs = {self.key_attribute_name: key}
        value = self.object_constructor(**kwargs)
        self[key] = value
        return value


@dataclass(order=True, unsafe_hash=True)
class FortigateObjectGroup:
    name: str


class FortigateIcmpService(FortigateObjectGroup):
    def __init__(self, name: str, icmp_types: list[int]):
        self.name = f'{name}-icmp'
        self.icmp_types = [str(i) for i in icmp_types]

    def __str__(self) -> str:
        """Return string representation of Fortigate service."""
        output = []
        output.append(f'    edit {self.name}')
        output.append(f'        set protocol ICMP')
        output.append(f'        set icmptype {" ".join(self.icmp_types)}')
        output.append(f'    next')
        return '\n'.join(output)


class FortigateIPService(FortigateObjectGroup):
    def __init__(
        self,
        name: str,
        source_port: list[tuple[int, int]],
        destination_port: list[tuple[int, int]],
        protocols: list[str],
    ):
        self.name = name
        # Adding a suffix to the name to avoid conflicts with term names
        self.tcp_port_range = None
        self.udp_port_range = None
        for proto in protocols:
            if proto == 'tcp':
                self.tcp_port_range = GenerateFortinetServiceString(source_port, destination_port)
            elif proto == 'udp':
                self.udp_port_range = GenerateFortinetServiceString(source_port, destination_port)
            elif proto == 'icmp':
                # ICMP is supported in FortigateIcmpService
                pass
            else:
                raise ValueError(f"Unsupported protocol: {proto}")

    def __str__(self) -> str:
        """Return string representation of Fortigate service."""
        output = []
        output.append(f'    edit {self.name}')
        if self.tcp_port_range:
            output.append(f'        set tcp-portrange {self.tcp_port_range}')
        if self.udp_port_range:
            output.append(f'        set udp-portrange {self.udp_port_range}')
        output.append('    next')
        return '\n'.join(output)


class FortinetAddress(FortigateObjectGroup):
    def __init__(self, name: str, ip: nacaddr.IP):
        """
        Initializes the FortinetAddress object.
        Args:
            name: The name of the address.
            ip: The IP address object.
        """
        self.name = name
        self.ip = ip

    def __str__(self) -> str:
        """Return string representation of Fortigate address."""
        output = []
        output.append(f'    edit {self.name}')
        if self.ip.version == 6:
            output.append(f'        set type ipprefix')
            output.append(f'        set ip6 {str(self.ip)}')
        else:
            output.append(f'        set type ipmask')
            output.append(
                f'        set subnet {str(self.ip.network_address)} {str(self.ip.netmask)}'
            )
        output.append('    next')
        return '\n'.join(output)


class FortigateExcludeGroup(FortigateObjectGroup):
    def __init__(self, name: str, members: list[str], exclude: list[str]):
        """
        Initializes the FortigateExcludeGroup object.
        Args:
            name: The name of the address group.
        """
        super().__init__(name=name)
        if not members:
            self._members_cfg_list = [FORTIGATE_ADDRESS_NONE]
        else:
            self._members_cfg_list = [f'"{m}"' for m in sorted(set(members))]
        if not exclude:
            self._exclude_cfg_list = [FORTIGATE_ADDRESS_NONE]
        else:
            self._exclude_cfg_list = [f'"{e}"' for e in sorted(set(exclude))]

    def __str__(self):
        output = []
        output.append(f'    edit {self.name}')
        output.append(f'        set member {" ".join(self._members_cfg_list)}')
        output.append(f'        set exclude enable')
        output.append(f'        set exclude-member {" ".join(self._exclude_cfg_list)}')
        output.append(f'    next')
        return '\n'.join(output)


class FortigateAddressGroup(FortigateObjectGroup):
    def __init__(self, name: str):
        """
        Initializes the FortigateAddressGroup object.
        Args:
            name: The name of the address group.
        """
        self.name = name
        self._cached_fortigate_addrs: list[FortinetAddress] = []
        self._members = set()
        self._is_dirty = True

    def AddMember(self, ip: nacaddr.IP) -> None:
        """Add an IP address to the group."""
        if ip not in self._members:
            self._members.add(ip)
            self._is_dirty = True

    @property
    def fortigate_addrs(self) -> list[FortinetAddress]:
        if self._is_dirty or not self._cached_fortigate_addrs:
            self._cached_fortigate_addrs = []
            sorted_ips = sorted(list(self._members))
            for i, ip_obj in enumerate(sorted_ips):
                address_object_name = f"{self.name}_{i}"
                self._cached_fortigate_addrs.append(
                    FortinetAddress(name=address_object_name, ip=ip_obj)
                )
            self._is_dirty = False
        return self._cached_fortigate_addrs

    def __contains__(self, item: Any) -> bool:
        """Check if an item is in the address group."""
        return item in self._members

    def __str__(self) -> str:
        """Return string representation of Fortigate address."""
        output = []
        current_fortigate_addrs = self.fortigate_addrs
        if not current_fortigate_addrs:  # Don't render empty groups
            return ""
        output.append(f'    edit {self.name}')
        member_names = [f'"{addr.name}"' for addr in current_fortigate_addrs]
        output.append(f'        set member {" ".join(member_names)}')
        output.append(f'    next')
        return '\n'.join(output)

    def __len__(self):
        return len(self._members)

    def __iter__(self):
        return iter(self.fortigate_addrs)


class Term(aclgenerator.Term):
    SADDR_V4 = 'srcaddr'
    DADDR_V4 = 'dstaddr'
    SADDR_V6 = 'srcaddr6'
    DADDR_V6 = 'dstaddr6'

    def __init__(
        self,
        term: policy.Term,
        source_interface: str,
        destination_interface: str,
        addressgroups: FortigateDefaultDictionary,
        addressgroups_v6: FortigateDefaultDictionary,
        address_family: str,
    ):
        """
        Initializes the Term object.
        Args:
            term: The term object from the policy.
            source_interface: The source interface for the term.
            destination_interface: The destination interface for the term.
        """
        super().__init__(term)
        self.term = term

        # Validate self.term.action
        if not isinstance(self.term.action, list):
            raise TypeError("term.action must be a list.")
        if len(self.term.action) != 1:
            raise ValueError("term.action must contain exactly one element.")
        if self.term.action[0] not in SUPPORTED_ACTIONS:
            raise ValueError("term.action must contain only 'accept' or 'deny'.")

        self.source_interface = source_interface
        self.destination_interface = destination_interface
        self.address_family = address_family

        self.address_groups = addressgroups
        self.address_groups_v6 = addressgroups_v6
        self.services = []

        self._TranslateAddresses(term.source_address)
        self._TranslateAddresses(term.destination_address)
        self._TranslateAddresses(term.source_address_exclude)
        self._TranslateAddresses(term.destination_address_exclude)
        if term.source_address_exclude:
            self._TranslateExcludes(
                f"{self.term.name}-source", term.source_address, term.source_address_exclude
            )
        if term.destination_address_exclude:
            self._TranslateExcludes(
                f"{self.term.name}-destination",
                term.destination_address,
                term.destination_address_exclude,
            )

        if term.destination_port or term.source_port:
            service = FortigateIPService(
                term.name, term.source_port, term.destination_port, term.protocol
            )
            self.services.append(service)
        if term.icmp_type:
            service = FortigateIcmpService(
                term.name, self.NormalizeIcmpTypes(term.icmp_type, term.protocol, 4)
            )
            self.services.append(service)
        # Logging defaults to UTM on fortigate
        # Set those defaults and then if logging is enabled, check options for
        # deviations from defaults.
        self.logtraffic = 'utm'
        self.logtraffic_start = 'disabled'
        if self.term.logging:
            if all(opt in self.term.option for opt in FORTIGATE_LOG_TRAFFIC_VALUES):
                raise ValueError(
                    "Multiple logtraffic values cannot be used together. "
                    f"Valid values are: {', '.join(FORTIGATE_LOG_TRAFFIC_VALUES)}"
                )
            # logtraffic options.
            if 'log_traffic_mode_all' in self.term.option:
                self.logtraffic = 'log_traffic_mode_all'
            elif 'log_traffic_mode_disable' in self.term.option:
                self.logtraffic = 'log_traffic_mode_disable'

            # logtraffic-start options.
            if 'log_traffic_start_session' in self.term.option:
                self.logtraffic_start = 'log_traffic_start_session'

    def _TranslateExcludes(
        self, base_name: str, addrs: list[nacaddr.IP], excludes: list[nacaddr.IP]
    ) -> None:
        member_tokens_v4 = [i.token for i in addrs if i.version == 4]
        member_tokens_v6 = [i.token for i in addrs if i.version == 6]
        exclude_tokens_v4 = [i.token for i in excludes if i.version == 4]
        exclude_tokens_v6 = [i.token for i in excludes if i.version == 6]

        # Default members to "none" if addrs is empty but excludes exist
        effective_members_v4 = member_tokens_v4 if member_tokens_v4 else [FORTIGATE_ADDRESS_NONE]
        effective_members_v6 = member_tokens_v6 if member_tokens_v6 else [FORTIGATE_ADDRESS_NONE]

        if exclude_tokens_v4:  # Only create if there are actual v4 excludes
            v4_exclude_group_name = f"{base_name}"  # Differentiated name
            exclude_addrgrp_v4 = FortigateExcludeGroup(
                v4_exclude_group_name, effective_members_v4, exclude_tokens_v4
            )
            self.address_groups[v4_exclude_group_name] = exclude_addrgrp_v4
        if exclude_tokens_v6:  # Only create if there are actual v6 excludes
            v6_exclude_group_name = f"{base_name}"  # Differentiated name
            exclude_addrgrp_v6 = FortigateExcludeGroup(
                v6_exclude_group_name, effective_members_v6, exclude_tokens_v6
            )
            self.address_groups_v6[v6_exclude_group_name] = exclude_addrgrp_v6

    def _TranslateAddresses(self, addrs: list[nacaddr.IP]) -> None:
        """Inserts tokens into versioned addressbook and sets their members to the IPs."""
        for addr in addrs:
            if addr.version == 4:
                if self.address_family == 'inet6':
                    continue
                self.address_groups[addr.token].AddMember(addr)
            elif addr.version == 6:
                if self.address_family == 'inet':
                    continue
                self.address_groups_v6[addr.token].AddMember(addr)
            else:
                raise ValueError(f"Unsupported address version: {addr.version}")

    def _SetStrLogging(self, output):
        if self.term.logging:
            if self.logtraffic == 'log_traffic_mode_all':
                output.append(f'        set logtraffic all')
            elif self.logtraffic == 'log_traffic_mode_disable':
                output.append(f'        set logtraffic disable')
            if self.logtraffic_start == 'log_traffic_start_session':
                output.append(f'        set logtraffic-start enable')
        else:
            output.append(f'        set logtraffic disable')
        return output

    def _SetStrInterfaces(self, output):
        output.append(f'        set srcintf "{self.source_interface}"')
        output.append(f'        set dstintf "{self.destination_interface}"')
        return output

    def __str__(self):
        """Return string representation of Fortigate term."""
        output = []
        output.append(f'        set name "{self.term.name}"')
        output = self._SetStrInterfaces(output)
        if self.term.action[0] != 'deny':
            output.append(f'        set action {self.term.action[0]}')
        # Collecting all tokens into bins of src/dst and IP version.
        source_tokens_v4 = set()
        source_tokens_v6 = set()
        for i in self.term.source_address:
            if i.version == 4:
                source_tokens_v4.add(f'"{i.token}"')
            elif i.version == 6:
                source_tokens_v6.add(f'"{i.token}"')
        destination_tokens_v4 = set()
        destination_tokens_v6 = set()
        for i in self.term.destination_address:
            if i.version == 4:
                destination_tokens_v4.add(f'"{i.token}"')
            elif i.version == 6:
                destination_tokens_v6.add(f'"{i.token}"')

        # Sorting tokens
        source_tokens_v4 = sorted(source_tokens_v4)
        source_tokens_v6 = sorted(source_tokens_v6)
        destination_tokens_v4 = sorted(destination_tokens_v4)
        destination_tokens_v6 = sorted(destination_tokens_v6)

        has_v4_source_elements = bool(
            source_tokens_v4 or any(ex.version == 4 for ex in self.term.source_address_exclude)
        )
        has_v6_source_elements = bool(
            source_tokens_v6 or any(ex.version == 6 for ex in self.term.source_address_exclude)
        )
        has_v4_destination_elements = bool(
            destination_tokens_v4
            or any(ex.version == 4 for ex in self.term.destination_address_exclude)
        )
        has_v6_destination_elements = bool(
            destination_tokens_v6
            or any(ex.version == 6 for ex in self.term.destination_address_exclude)
        )

        generate_ipv4_policy = False
        generate_ipv6_policy = False

        if self.address_family == 'inet':
            generate_ipv4_policy = True
        elif self.address_family == 'inet6':
            generate_ipv6_policy = True
        elif self.address_family == 'mixed':
            # If there are any v4 elements (source or dest, include or exclude)
            # OR if there are no v6 elements at all (meaning v4 is the only option or it's truly 'any/any')
            if (has_v4_source_elements or has_v4_destination_elements) or not (
                has_v6_source_elements or has_v6_destination_elements
            ):
                generate_ipv4_policy = True

            # If there are any v6 elements (source or dest, include or exclude)
            # OR if there are no v4 elements at all
            if (has_v6_source_elements or has_v6_destination_elements) or not (
                has_v4_source_elements or has_v4_destination_elements
            ):
                generate_ipv6_policy = True

            # If both are true but one side has no elements, turn off the other.
            # e.g. if only v6 exclude exists, don't generate an empty/all v4 rule.
            if generate_ipv4_policy and generate_ipv6_policy:
                if not (has_v4_source_elements or has_v4_destination_elements):
                    generate_ipv4_policy = False
                if not (has_v6_source_elements or has_v6_destination_elements):
                    generate_ipv6_policy = False

            # If mixed and absolutely no address specifications (includes or excludes)
            # default to generating for both (Fortigate 'all'/'all')
            if not (
                has_v4_source_elements
                or has_v4_destination_elements
                or has_v6_source_elements
                or has_v6_destination_elements
            ):
                generate_ipv4_policy = True
                generate_ipv6_policy = True

        # Default actions
        src_v4_effective_val = FORTIGATE_ADDRESS_ALL
        if not has_v4_source_elements and self.term.source_address:
            src_v4_effective_val = FORTIGATE_ADDRESS_NONE

        dst_v4_effective_val = FORTIGATE_ADDRESS_ALL
        if not has_v4_destination_elements and self.term.destination_address:
            dst_v4_effective_val = FORTIGATE_ADDRESS_NONE

        src_v6_effective_val = FORTIGATE_ADDRESS_ALL
        if not has_v6_source_elements and self.term.source_address:
            src_v6_effective_val = FORTIGATE_ADDRESS_NONE

        dst_v6_effective_val = FORTIGATE_ADDRESS_ALL
        if not has_v6_destination_elements and self.term.destination_address:
            dst_v6_effective_val = FORTIGATE_ADDRESS_NONE

        # Generate IPv4 address groups from tokens or use default action if empty
        if generate_ipv4_policy:
            # Check if there's a v4-specific source exclude group
            source_v4_exclude_exists = any(
                ex.version == 4 for ex in self.term.source_address_exclude
            )
            if source_v4_exclude_exists:
                # Ensure the _TranslateExcludes created a V4 group for this term.
                # The name needs to be consistently generated.
                src_v4_to_set = f'"{self.term.name}-source"'  # Assumes _TranslateExcludes uses this name for the v4 version
            elif source_tokens_v4:
                src_v4_to_set = " ".join(source_tokens_v4)
            else:
                src_v4_to_set = f'"{src_v4_effective_val}"'  # Use effective default (all or none)

            destination_v4_exclude_exists = any(
                ex.version == 4 for ex in self.term.destination_address_exclude
            )
            if destination_v4_exclude_exists:
                dst_v4_to_set = f'"{self.term.name}-destination"'
            elif destination_tokens_v4:
                dst_v4_to_set = " ".join(destination_tokens_v4)
            else:
                dst_v4_to_set = f'"{dst_v4_effective_val}"'

            output.append(f'        set {self.SADDR_V4} {src_v4_to_set}')
            output.append(f'        set {self.DADDR_V4} {dst_v4_to_set}')

        # Generate IPv6 address groups
        if generate_ipv6_policy:
            source_v6_exclude_exists = any(
                ex.version == 6 for ex in self.term.source_address_exclude
            )
            if source_v6_exclude_exists:
                src_v6_to_set = (
                    f'"{self.term.name}-source"'  # This name might need adjustment if it clashes
                )
            elif source_tokens_v6:
                src_v6_to_set = " ".join(source_tokens_v6)
            else:
                src_v6_to_set = f'"{src_v6_effective_val}"'

            destination_v6_exclude_exists = any(
                ex.version == 6 for ex in self.term.destination_address_exclude
            )
            if destination_v6_exclude_exists:
                dst_v6_to_set = f'"{self.term.name}-destination"'
            elif destination_tokens_v6:
                dst_v6_to_set = " ".join(destination_tokens_v6)
            else:
                dst_v6_to_set = f'"{dst_v6_effective_val}"'

            output.append(f'        set {self.SADDR_V6} {src_v6_to_set}')
            output.append(f'        set {self.DADDR_V6} {dst_v6_to_set}')
        output.append(f'        set schedule {FORTIGATE_SCHEDULE_ALWAYS}')
        services = [f'"{service.name}"' for service in self.services]
        if not services:
            services = [f'"{FORTIGATE_SERVICES_ALL}"']
        output.append(f'        set service {"".join(services)}')

        # Multiple logging types are available, check if logging is true and then print relevant log statements.
        output = self._SetStrLogging(output)
        # Comments have a limited length. Add Owner to the comment and then join the term comment.
        # If the comment is too long warn the user and truncate.
        comment = []
        if self.term.owner:
            comment.append(f'Owner: {self.term.owner}')
        if self.term.comment:
            comment.append(' '.join(self.term.comment))
        comment = '\n'.join(comment)
        if len(comment) >= FORTIGATE_COMMENT_LIMIT:
            logging.warning(
                f"Fortigate comment cannot be longer than {FORTIGATE_COMMENT_LIMIT} characters, length was {len(comment)}, truncating comment."
            )
            comment = comment[:FORTIGATE_COMMENT_LIMIT]
        if comment:
            output.append(f'        set comments "{comment}"')
        return '\n'.join(output)


class LocalInTerm(Term):
    SADDR_V4 = 'srcaddr'
    DADDR_V4 = 'dstaddr'
    SADDR_V6 = 'srcaddr'
    DADDR_V6 = 'dstaddr'

    def _SetStrLogging(self, output):
        # No logging for Local in terms.
        return output

    def _SetStrInterfaces(self, output):
        output.append(f'        set intf "{self.source_interface}"')
        return output


class Fortigate(aclgenerator.ACLGenerator):
    """Fortigate ACL generator."""

    _PLATFORM = 'fortigate'
    SUFFIX = '.fgacl'
    _SUPPORTED_AF = frozenset(('inet', 'inet6', 'mixed'))

    def __init__(self, name, description):
        """Initialize Fortigate ACL generator.
        Args:
          name: Name of the ACL.
          description: Description of the ACL.
        """
        self.services = []
        self.terms = []
        self.local_in_terms_v4 = []
        self.local_in_terms_v6 = []
        self.term_names = set()
        self.address_groups = FortigateDefaultDictionary(FortigateAddressGroup, 'name')
        self.address_groups_v6 = FortigateDefaultDictionary(FortigateAddressGroup, 'name')
        super().__init__(name, description)

    @staticmethod
    def _SupportedOptions(config: MutableMapping) -> list[AbstractOption]:
        return [
            ArbitraryValueOption("from-zone", config),
            ArbitraryValueOption("to-zone", config),
            BooleanKeywordOption("mixed", config),
            BooleanKeywordOption("inet6", config),
            BooleanKeywordOption("inet", config),
        ]

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()
        supported_tokens.remove('stateless_reply')
        supported_sub_tokens['action'] = SUPPORTED_ACTIONS
        supported_sub_tokens['option'] = (
            FORTIGATE_LOG_TRAFFIC_VALUES | FORTIGATE_LOG_TRAFFIC_START_VALUES
        )
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)

            filter_config = ProcessOptions(
                self._SupportedOptions,
                filter_options,
            )

            if "from-zone" not in filter_config or "to-zone" not in filter_config:
                raise ValueError(
                    f'Fortigate requires that at least from-zone and to-zone interfaces/zone names be specified, found: {filter_options}'
                )
            source_interface = filter_config["from-zone"]
            destination_interface = filter_config["to-zone"]

            af = self._SUPPORTED_AF.intersection(filter_config.keys())
            if len(af) > 1:
                raise ValueError(
                    f"Only one address family is supported (inet, inet6, mixed) but found: {', '.join(af)}"
                )
            if destination_interface == "local-in-policy":
                if not af:
                    raise ValueError(
                        "Address family must be specified as 'inet' or 'inet6' for local-in-policy."
                    )
                af = next(iter(af))
                if af not in ('inet', 'inet6'):
                    raise ValueError("Only 'inet' or 'inet6' are supported for local-in-policy.")

            else:
                if not af:
                    af = 'mixed'
                else:
                    af = next(iter(af))
            TermClass = LocalInTerm if destination_interface == "local-in-policy" else Term

            for term in terms:
                fortigate_term = TermClass(
                    term,
                    source_interface,
                    destination_interface,
                    self.address_groups,
                    self.address_groups_v6,
                    af,
                )
                self.services.extend(fortigate_term.services)
                self.term_names.add(term.name)
                if destination_interface == "local-in-policy":
                    if af == "inet":
                        self.local_in_terms_v4.append(fortigate_term)
                    elif af == "inet6":
                        self.local_in_terms_v6.append(fortigate_term)
                    else:
                        raise ValueError(
                            "local-in-policy must specify 'inet' or 'inet6' as address family."
                        )
                else:
                    self.terms.append(fortigate_term)
        self._ValidateNoDuplicateTerms(self.terms)
        self._ValidateNoDuplicateTerms(self.local_in_terms_v4)
        self._ValidateNoDuplicateTerms(self.local_in_terms_v6)

    def _ValidateNoDuplicateTerms(self, term_list):
        seen = set()
        for term in term_list:
            if term.term.name in seen:
                raise ValueError(
                    f"Duplicate term name '{term.term.name}' found in the same term list."
                )
            seen.add(term.term.name)

    def _RenderPolicySection(self, output, section_name, terms):
        if terms:
            output.append(f'config firewall {section_name}')
            counter = 1
            for term in terms:
                output.append(f'    edit {counter}')
                output.append(str(term))
                counter += 1
                output.append('    next')
            output.append('end')
        return output

    def __str__(self) -> str:
        """Return string representation of Fortigate ACL."""
        output = []

        # IPv4
        if self.address_groups.values():
            # IPv4 Addresses
            output.append('config firewall address')
            for group_name in sorted(self.address_groups):
                group = self.address_groups[group_name]
                # Don't process excludes in addresses
                if isinstance(group, FortigateAddressGroup):
                    for addr in sorted(self.address_groups[group_name].fortigate_addrs):
                        output.append(str(addr))
            output.append('end')
            # Address Groups
            output.append('config firewall addrgrp')
            for addr_group in sorted(self.address_groups.values(), key=lambda item: item.name):
                output.append(str(addr_group))
            output.append('end')

        # IPv6 Space
        if self.address_groups_v6.values():
            # IPv6 Addresses
            output.append('config firewall address6')
            for group_name in sorted(self.address_groups_v6):
                group = self.address_groups_v6[group_name]
                # Don't process excludes in addresses
                if isinstance(group, FortigateAddressGroup):
                    for addr in sorted(self.address_groups_v6[group_name].fortigate_addrs):
                        output.append(str(addr))
            output.append('end')
            # Address Groups
            output.append('config firewall addrgrp6')
            for addr_group in sorted(self.address_groups_v6.values(), key=lambda item: item.name):
                output.append(str(addr_group))
            output.append('end')

        # Services
        if self.services:
            output.append('config firewall service custom')
            for service in sorted(self.services):
                output.append(str(service))
            output.append('end')

        # Firewall Policies

        output = self._RenderPolicySection(output, 'policy', self.terms)
        output = self._RenderPolicySection(output, 'local-in-policy', self.local_in_terms_v4)
        output = self._RenderPolicySection(output, 'local-in-policy6', self.local_in_terms_v6)
        return '\n'.join(output)


def FormatFortinetPortRange(low: int, high: int) -> str:
    """Formats a single port range according to Fortinet syntax."""
    if not (1 <= low <= 65535 and 1 <= high <= 65535):
        raise ValueError(f"Ports must be between 1 and 65535. Got: low={low}, high={high}")
    if low > high:
        raise ValueError(f"Low port cannot be greater than high port. Got: low={low}, high={high}")
    if low == high:
        return str(low)
    else:
        return f"{low}-{high}"


def GenerateFortinetServiceString(
    source_ranges: list[tuple[int, int]], destination_ranges: list[tuple[int, int]]
) -> str:
    """
    Generates a Fortinet service port range string by combining destination and source ranges.

    Args:
        destination_ranges: A list of tuples, where each tuple is (low_port, high_port)
                            for destination ports.
        source_ranges: A list of tuples, where each tuple is (low_port, high_port)
                    for source ports.

    Returns:
        A space-separated string suitable for Fortinet's [tcp|udp]-portrange commands.
        Example: "80:80-90 80:100 120-121:80-90 120-121:100"
    """
    output_parts = []

    for dst_low, dst_high in destination_ranges:
        dest_str = FormatFortinetPortRange(dst_low, dst_high)
        if not source_ranges:
            # If no source ranges are specified, only list destination ranges
            # (Fortinet defaults source to 1-65535)
            output_parts.append(dest_str)
        else:
            # If source ranges ARE specified, combine each dest with each source
            for src_low, src_high in source_ranges:
                src_str = FormatFortinetPortRange(src_low, src_high)
                output_parts.append(f"{dest_str}:{src_str}")

    return " ".join(output_parts)



import re
from dataclasses import dataclass


@dataclass
class FQDN:
    """Contains a single fully qualified domain name associated with a token, parent_token and optional comment."""

    # https://regexr.com/3g5j0
    fqdn_re = re.compile(
        r'^(?!.*:\/\/)(?=.{1,255}$)((.{1,63}\.){1,127}(?![0-9]*$)[a-z0-9-]+\.?)$', re.IGNORECASE
    )
    fqdn: str
    text: str
    token: str
    parent_token: str

    def __init__(self, fqdn: str, token: str, comment: str = ''):
        if not self.fqdn_re.match(fqdn):
            raise ValueError(f"Not a valid FQDN: {fqdn}")
        self.fqdn = fqdn
        self.text = comment
        self.token = token
        self.parent_token = token



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Parse naming definition files.

Network access control applications use definition files which contain
information about networks and services.  This naming class
will provide an easy interface into using these definitions.

Sample usage with definition files contained in ./acl/defs:
    defs = Naming('acl/defs/')

    services =  defs.GetService('DNS')
      returns ['53/tcp', '53/udp', ...]

    networks = defs.GetNet('INTERNAL')
      returns a list of nacaddr.IPv4 object

The definition files are contained in a single directory and
may consist of multiple files ending in .net or .svc extensions,
indicating network or service definitions respectively.  The
format of the files consists of a 'token' value, followed by a
list of values and optional comments, such as:

INTERNAL = 10.0.0.0/8     # RFC-1918
           172.16.0.0/12  # RFC-1918
           192.168.0.0/16 # RFC-1918
or

DNS = 53/tcp
      53/udp

"""


import re
from pathlib import Path
from typing import Any, Optional, Union

import yaml
from absl import logging
from yaml import YAMLError

from aerleon.lib import nacaddr
from aerleon.lib import port as portlib
from aerleon.lib.fqdn import FQDN
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.yaml_loader import SpanSafeYamlLoader

DEF_TYPE_SERVICES = 'services'
DEF_TYPE_NETWORKS = 'networks'


class Error(Exception):
    """Create our own base error class to be inherited by other error classes."""


class NamespaceCollisionError(Error):
    """Used to report on duplicate symbol names found while parsing."""


class BadNetmaskTypeError(Error):
    """Deprecated."""


class NoDefinitionsError(Error):
    """Raised if no definitions are found."""


class ParseError(Error):
    """Raised if an error occurs during parsing."""


class UndefinedAddressError(Error):
    """Raised if an address is referenced but not defined."""


class UndefinedServiceError(Error):
    """Raised if a service is referenced but not defined."""


class UndefinedPortError(Error):
    """Raised if a port/protocol pair has not been defined."""


class UnexpectedDefinitionTypeError(Error):
    """An unexpected/unknown definition type was used."""


class NamingSyntaxError(Error):
    """A general syntax error for the definition."""


class DefinitionFileTypeError(Error):
    """Invalid Definition File"""


class EmptyDefinitionError(Error):
    """A token returned no results."""


# Consider making this span-oriented
# (file > line > (start_ch, end_ch))
class UserMessage:
    """A user-facing error message encountered during file processing.

    Users can be shown:
    * An error message only (user_message.message).
    * An error message with file / line / include stack (user_message.__repr__()).

    Attributes:
        message: The error message.
        filename: The name of the file in which this error or message originated.
        line: The line where this error or message originated.
        include_chain: If the error or message originated while processing an included
            file, include_chain will list the include file chain as a list of file/line tuples.
            The top-level file should be the first item in the list.
    """

    def __init__(
        self,
        message: str,
        *,
        filename: str,
        line: Optional[int] = None,
        include_chain: Optional[list[tuple[str, int]]] = None,
    ):
        self.message = message
        self.filename = filename
        self.line = line
        self.include_chain = include_chain

    def __str__(self) -> str:
        """Display user-facing error message with include chain (if present).

        e.g.
        Excessive recursion: include depth limit of 5 reached. File=include_1.pol-include.yaml, Line=3.
        Include stack:
        > File='policy_with_include.pol.yaml', Line=11 (Top Level)
        > File='include_1.pol-include.yaml', Line=3
        > File='include_1.pol-include.yaml', Line=3
        > File='include_1.pol-include.yaml', Line=3
        > File='include_1.pol-include.yaml', Line=3
        > File='include_1.pol-include.yaml', Line=3
        """  # noqa: E501
        error_context = f"{self.message} File={self.filename}"
        if self.line is not None:
            error_context += f", Line={self.line}"
        error_context += "."
        if self.include_chain is not None and len(self.include_chain) > 1:
            error_context += "\nInclude stack:"
            for i, (File, Line) in enumerate(self.include_chain):
                error_context += f"\n> File='{File}', Line={Line}"
                if i == 0:
                    error_context += " (Top Level)"
        return error_context

    def __repr__(self) -> str:
        return f"UserMessage(\"{str(self)}\")"


def is_yaml_suffix(suffix: str) -> bool:
    return suffix == '.yaml' or suffix == '.yml'


class _ItemUnit:
    """This class is a container for an index key and a list of associated values.

    An ItemUnit will contain the name of either a service or network group,
    and a list of the associated values separated by spaces.

    Attributes:
      name: A string representing a unique token value.
      items: a list of strings containing values for the token.
    """

    def __init__(self, symbol: str) -> None:
        self.name = symbol
        self.items = []


class Naming:
    """Object to hold naming objects from NETWORK and SERVICES definition files.

    Attributes:
       current_symbol: The current token being handled while parsing data.
       services: A collection of all of the current service item tokens.
       networks: A collection of all the current network item tokens.
       unseen_services: Undefined service entries.
       unseen_networks: Undefined network entries.
       port_re: Regular Expression matching valid port entries.
    """

    def __init__(
        self,
        naming_dir: Optional[str] = None,
        naming_file: Optional[str] = None,
        naming_type: Optional[str] = None,
    ) -> None:
        """Set the default values for a new Naming object.

        Args:
          naming_dir: A string containing a file path to the directory where
            definition files are located.
          naming_file: Optional. A string containing the file path to a specific
            defintion file. Only this file will be loaded if naming_file is given.
          naming_type: Optional. A string containing either 'service' or 'network'.
            This option is only needed if naming_file is provided and it refers to
            a non-YAML file.
        """
        self.current_symbol = None
        self.services = {}
        self.networks = {}
        self.fqdn = {}
        self.unseen_services = {}
        self.unseen_networks = {}
        self.port_re = re.compile(r'(^\d+-\d+|^\d+)\/\w+$|^[\w\d-]+$', re.IGNORECASE | re.DOTALL)
        self.token_re = re.compile(r'(^[-_A-Z0-9]+$)', re.IGNORECASE)

        if naming_file:
            file_path = Path(naming_dir).joinpath(naming_file)
            if is_yaml_suffix(file_path.suffix):
                if naming_type:
                    logging.warning('Naming object: ignoring unexpected naming_type.')

                with open(file_path) as file_handle:
                    self.ParseYaml(file_handle, file_path.name)
            elif naming_type:
                with open(file_path) as file_handle:
                    self._ParseFile(file_handle, naming_type)

        elif naming_dir:
            if naming_type:
                logging.warning('Naming object: ignoring unexpected naming_type.')

            self._Parse(naming_dir)
            self._CheckUnseen(DEF_TYPE_SERVICES)
            self._CheckUnseen(DEF_TYPE_NETWORKS)

    def _CheckUnseen(self, def_type: str):
        if def_type == DEF_TYPE_SERVICES:
            if self.unseen_services:
                raise UndefinedServiceError(
                    '%s %s'
                    % (
                        'The following tokens were nested as a values, but not defined',
                        self.unseen_services,
                    )
                )
        if def_type == DEF_TYPE_NETWORKS:
            if self.unseen_networks:
                raise UndefinedAddressError(
                    '%s %s'
                    % (
                        'The following tokens were nested as a values, but not defined',
                        self.unseen_networks,
                    )
                )

    def GetIpParents(self, query: str) -> list[str]:
        """Return network tokens that contain IP in query.

        Args:
          query: an ip string ('10.1.1.1') or nacaddr.IP object

        Returns:
          A sorted list of unique parent tokens.
        """
        base_parents = []
        recursive_parents = []
        # convert string to nacaddr, if arg is ipaddr then convert str() to nacaddr
        if not isinstance(query, nacaddr.IPv4) and not isinstance(query, nacaddr.IPv6):
            if query[:1].isdigit():
                query = nacaddr.IP(query)
        # Get parent token for an IP
        if isinstance(query, nacaddr.IPv4) or isinstance(query, nacaddr.IPv6):
            for token in self.networks:
                for item in self.networks[token].items:
                    item = item.split('#')[0].strip()
                    if not item[:1].isdigit():
                        continue
                    try:
                        supernet = nacaddr.IP(item, strict=False)
                        if supernet.supernet_of(query):
                            base_parents.append(token)
                    except ValueError:
                        # item was not an IP
                        pass
        # Get parent token for another token
        else:
            for token in self.networks:
                for item in self.networks[token].items:
                    item = item.split('#')[0].strip()
                    if item[:1].isalpha() and item == query:
                        base_parents.append(token)
        # look for nested tokens
        for bp in base_parents:
            done = False
            for token in self.networks:
                if bp in [item.split('#')[0].strip() for item in self.networks[token].items]:
                    # ignore IPs, only look at token values
                    if bp[:1].isalpha():
                        if bp not in recursive_parents:
                            recursive_parents.append(bp)
                            recursive_parents.extend(self.GetIpParents(bp))
                        done = True
            # if no nested tokens, just append value
            if not done:
                if bp[:1].isalpha() and bp not in recursive_parents:
                    recursive_parents.append(bp)
        return sorted(list(set(recursive_parents)))

    def GetServiceParents(self, query: str) -> list[str]:
        """Given a query token, return list of services definitions with that token.

        Args:
          query: a service token name.
        Returns:
          List of service definitions containing the token.
        """
        return self._GetParents(query, self.services)

    def GetNetParents(self, query: str) -> list[str]:
        """Given a query token, return list of network definitions with that token.

        Args:
          query: a network token name.
        Returns:
          A list of network definitions containing the token.
        """
        return self._GetParents(query, self.networks)

    def _GetParents(self, query: str, query_group: dict[str, _ItemUnit]) -> list[str]:
        """Given a naming item dict, return any tokens containing the value.

        Args:
          query: a service or token name, such as 53/tcp or DNS
          query_group: either services or networks dict

        Returns:
          Returns a list of definitions containing the token in desired group.
        """
        base_parents = []
        recursive_parents = []
        # collect list of tokens containing query
        for token in query_group:
            if query in [item.split('#')[0].strip() for item in query_group[token].items]:
                base_parents.append(token)
        if not base_parents:
            return []
        # iterate through tokens containing query, doing recursion if necessary
        for bp in base_parents:
            for token in query_group:
                if bp in query_group[token].items and bp not in recursive_parents:
                    recursive_parents.append(bp)
                    recursive_parents.extend(self._GetParents(bp, query_group))
            if bp not in recursive_parents:
                recursive_parents.append(bp)
        return recursive_parents

    def GetNetChildren(self, query: str) -> list[str]:
        """Given a query token, return list of network definitions tokens within provided token.

        This will only return children, not descendants of provided token.

        Args:
          query: a network token name.

        Returns:
          A list of network definitions tokens within this token.
        """
        return self._GetChildren(query, self.networks)

    def _GetChildren(self, query: str, query_group: dict[str, _ItemUnit]) -> list[str]:
        """Given a naming item dict, return tokens (not IPs) contained within this value.

        Args:
          query: a token name
          query_group: networks dict

        Returns:
          Returns a list of definitions tokens within (children) target token.
        """

        children = []
        if query in query_group:
            for item in query_group[query].items:
                child = item.split('#')[0].strip()

                # Determine if item a token, then it's a child
                if not self._IsIpFormat(child):
                    children.append(child)

        return children

    def _IsIpFormat(self, item: str) -> bool:
        """Helper function for _GetChildren to detect if string is IP format.

        Args:
          item: string either a IP or token.
        Returns:
          True if string is a IP
          False if string is not a IP
        """
        try:
            item = item.strip()
            nacaddr.IP(item, strict=False)
            return True
        except ValueError:
            return False

    def GetServiceNames(self) -> list[str]:
        """Returns the list of all known service names."""
        return list(self.services.keys())

    def GetService(self, query: str) -> list[str]:
        """Given a service name, return a list of associated ports and protocols.

        Args:
          query: Service name symbol or token.

        Returns:
          A list of service values such as ['80/tcp', '443/tcp', '161/udp', ...]

        Raises:
          UndefinedServiceError: If the service name isn't defined.
        """
        expandset = set()
        already_done = set()
        data = []
        service_name = ''
        data = query.split('#')  # Get the token keyword and remove any comment
        service_name = data[0].split()[0]  # strip and cast from list to string
        if service_name not in self.services:
            raise UndefinedServiceError(f'\nNo such service: {query}')

        already_done.add(service_name)

        for next_item in self.services[service_name].items:
            # Remove any trailing comment.
            service = next_item.split('#')[0].strip()
            # Recognized token, not a value.
            if '/' not in service:
                # Make sure we are not descending into recursion hell.
                if service not in already_done:
                    already_done.add(service)
                    try:
                        expandset.update(self.GetService(service))
                    except UndefinedServiceError as e:
                        # One of the services in query is undefined, refine the error msg.
                        raise UndefinedServiceError(f'{e} (in {query})')
            else:
                expandset.add(service)
        return sorted(expandset)

    def GetPortParents(self, query: str, proto: str) -> list[str]:
        """Returns a list of all service tokens containing the port/protocol pair.

        Args:
            query: port number ('22') as str
            proto: protocol name ('tcp') as str

        Returns:
            A list of service tokens: ['SSH', 'HTTPS']

        Raises:
          UndefinedPortError: If the port/protocol pair isn't used in any
          service tokens.
        """
        # turn the given port and protocol into a PortProtocolPair object
        given_ppp = portlib.PPP(f"{query}/{proto}")
        base_parents = []
        matches = set()
        # check each service token to see if it's a PPP or a nested group.
        # if it's a PPP, see if there's a match with given_ppp
        # otherwise, add nested group to a list to recurisvely check later.
        # if there's no match, do nothing.
        for service_token in self.services:
            for port_child in self.services[service_token].items:
                ppp = portlib.PPP(port_child)
                # check for exact match
                if ppp.is_single_port and ppp == given_ppp:
                    matches.add(service_token)
                # check if it's within ppp's port range
                elif ppp.is_range and given_ppp in ppp:
                    matches.add(service_token)
                # if it's a nested token, add to a list to recurisvely
                # check later.
                elif ppp.nested:
                    if service_token not in base_parents:
                        base_parents.append(service_token)
        # break down the nested service tokens into PPP objects and check
        # against given_ppp
        for bp in base_parents:
            for port_child in self.GetService(bp):
                ppp = portlib.PPP(port_child)
                # check for exact match
                if ppp.is_single_port and ppp == given_ppp:
                    matches.add(bp)
                # check if it's within ppp's port range
                elif ppp.is_range and given_ppp in ppp:
                    matches.add(bp)
        # error if the port/protocol pair is not found.
        if not matches:
            raise UndefinedPortError(f'{query}/{proto} is not found in any service tokens')
        return sorted(matches)

    def GetServiceByProto(self, query: str, proto: str) -> list[str]:
        """Given a service name, return list of ports in the service by protocol.

        Args:
          query: Service name to lookup.
          proto: A particular protocol to restrict results by, such as 'tcp'.

        Returns:
          A list of service values of type 'proto', such as ['80', '443', ...]

        Raises:
          UndefinedServiceError: If the service name isn't defined.
        """
        services_set = set()
        proto = proto.upper()
        data = []
        servicename = ''
        data = query.split('#')  # Get the token keyword and remove any comment
        servicename = data[0].split()[0]  # strip and cast from list to string
        if servicename not in self.services:
            raise UndefinedServiceError('{} {}'.format('\nNo such service,', servicename))

        for service in self.GetService(servicename):
            if service and '/' in service:
                parts = service.split('/')
                if parts[1].upper() == proto:
                    services_set.add(parts[0])
        return sorted(services_set)

    def GetFQDN(self, query: str) -> list[str]:
        """Expand a network token into a list of FQDN objects.

        Args:
          query: Network definition token. May include comment text

        Returns:
          List of nacaddr.IPv4 or nacaddr.IPv6 objects

        Raises:
          UndefinedAddressError: Network token not defined
          EmptyDefinitionError: No FQDN values found for this network token
        """
        results = self._GetFQDN(query)
        if len(results) == 0:
            raise EmptyDefinitionError(f"No FQDN values found for network: {query}")
        return results

    def _GetFQDN(self, query: str, level=0) -> list[str]:
        returnlist = []
        data = query.split('#')
        token = data[0].split()[0]
        if token not in self.fqdn:
            raise UndefinedAddressError(f'UNDEFINED: {token}')
        for i in self.fqdn[token].items:
            comment = ''
            if i.find('#') > -1:
                (name, comment) = i.split('#', 1)
            else:
                name = i

            name = name.strip()
            if self.token_re.match(name):
                returnlist.extend(self._GetFQDN(name))
            else:
                try:
                    fqdn = FQDN(name, token, comment)
                    fqdn.text = comment.lstrip()
                    fqdn.token = token
                    returnlist.append(fqdn)
                except ValueError:
                    pass
        for i in returnlist:
            i.parent_token = token
        return returnlist

    def GetNetAddr(self, query: str) -> list[Union[IPv4, IPv6]]:
        """Alias of Naming.GetNet"""
        return self.GetNet(query)

    def GetNet(self, query: str) -> list[Union[IPv4, IPv6]]:
        """Expand a network token into a list of nacaddr.IPv4 or nacaddr.IPv6 objects.

        Args:
          query: Network definition token. May include comment text.

        Returns:
          List of nacaddr.IPv4 or nacaddr.IPv6 objects

        Raises:
          UndefinedAddressError: Network token not defined
          EmptyDefinitionError: No IP address values found for this network token
        """
        results = self._GetNet(query)
        if len(results) == 0:
            raise EmptyDefinitionError(f"No IP addresses found for network: {query}")
        return results

    def _GetNet(self, query: str) -> list[Union[IPv4, IPv6]]:
        returnlist = []
        data = query.split('#')
        token = data[0].split()[0]
        if token not in self.networks:
            raise UndefinedAddressError(f'UNDEFINED: {token}')

        for i in self.networks[token].items:
            comment = ''
            if i.find('#') > -1:
                (net, comment) = i.split('#', 1)
            else:
                net = i

            net = net.strip()
            if self.token_re.match(net):
                returnlist.extend(self.GetNet(net))
            else:
                try:
                    addr = nacaddr.IP(net, strict=False)
                    addr.text = comment.lstrip()
                    addr.token = token
                    returnlist.append(addr)
                except ValueError:
                    pass
        for i in returnlist:
            i.parent_token = token
        return returnlist

    def _Parse(self, definitions_directory: str) -> None:
        """Parse files for tokens and values.

        Given a directory name, grab all the appropriate files in that
        directory and parse them for definitions.

        Args:
          defdirectory: Path to directory containing definition files.
          def_type: Type of definitions to parse

        Raises:
          NoDefinitionsError: if no definitions are found.
        """
        file_def_type = {
            '.net': DEF_TYPE_NETWORKS,
            '.svc': DEF_TYPE_SERVICES,
            '.yaml': 'yaml',
            '.yml': 'yaml',
        }

        for path in Path(definitions_directory).iterdir():
            def_type = file_def_type.get(path.suffix)

            if not def_type:
                continue

            try:
                with open(path) as file:
                    if def_type == 'yaml':
                        self.ParseYaml(file, path.name)
                    else:
                        self._ParseFile(file, def_type)

            except OSError as error_info:
                raise NoDefinitionsError(f'{error_info}')

    def _ParseFile(self, file_handle: list[str], def_type: str) -> None:
        for line in file_handle:
            self._ParseLine(line, def_type)

    def ParseServiceList(self, data: list[str]) -> None:
        """Take an array of service data and import into class.

        This method allows us to pass an array of data that contains service
        definitions that are appended to any definitions read from files.

        Args:
          data: array of text lines containing service definitions.
        """
        for line in data:
            self._ParseLine(line, DEF_TYPE_SERVICES)

    def ParseNetworkList(self, data: list[str]) -> None:
        """Take an array of network data and import into class.

        This method allows us to pass an array of data that contains network
        definitions that are appended to any definitions read from files.

        Args:
          data: array of text lines containing net definitions.

        """
        for line in data:
            self._ParseLine(line, DEF_TYPE_NETWORKS)

    def _ParseLine(self, line: str, definition_type: str) -> None:
        """Parse a single line of a service definition file.

        This routine is used to parse a single line of a service
        definition file, building a list of 'self.services' objects
        as each line of the file is iterated through.

        Args:
          line: A single line from a service definition files.
          definition_type: Either 'networks' or 'services'

        Raises:
          UnexpectedDefinitionTypeError: called with unexpected type of definitions.
          NamespaceCollisionError: when overlapping tokens are found.
          ParseError: If errors occur
          NamingSyntaxError: Syntax error parsing config.
        """

        if definition_type not in ['services', 'networks']:
            raise UnexpectedDefinitionTypeError(
                f'Received an unexpected definition type: {definition_type}'
            )
        line = line.strip()
        if not line or line.startswith('#'):  # Skip comments and blanks.
            return
        comment = ''
        if line.find('#') > -1:  # if there is a comment, save it
            (line, comment) = line.split('#', 1)
        line_parts = line.split('=')  # Split on var = val lines.
        # the value field still has the comment at this point
        # If there was '=', then do var and value
        if len(line_parts) > 1:
            current_symbol = line_parts[0].strip()  # varname left of '='
            if not self.token_re.match(current_symbol):
                logging.info(
                    '\nService name does not match recommended criteria: %s\nOnly A-Z, a-z, 0-9, -, and _ allowed'
                    % current_symbol
                )  # TODO(jb) This error is incorrect when current_symbol is a network name
            self.current_symbol = current_symbol
            if definition_type == DEF_TYPE_SERVICES:
                for port in line_parts[1].strip().split():
                    if not self.port_re.match(port):
                        raise NamingSyntaxError(f'The following line has a syntax error: {line}')
                if self.current_symbol in self.services:
                    raise NamespaceCollisionError(
                        '%s %s'
                        % ('\nMultiple definitions found for service: ', self.current_symbol)
                    )
            elif definition_type == DEF_TYPE_NETWORKS:
                if self.current_symbol in self.networks:
                    raise NamespaceCollisionError(
                        '%s %s'
                        % ('\nMultiple definitions found for network: ', self.current_symbol)
                    )

            self.unit = _ItemUnit(self.current_symbol)
            if definition_type == DEF_TYPE_SERVICES:
                self.services[self.current_symbol] = self.unit
                # unseen_services is a list of service TOKENS found in the values
                # of newly defined services, but not previously defined themselves.
                # When we define a new service, we should remove it (if it exists)
                # from the list of unseen_services.
                if self.current_symbol in self.unseen_services:
                    self.unseen_services.pop(self.current_symbol)
            elif definition_type == DEF_TYPE_NETWORKS:
                self.networks[self.current_symbol] = self.unit
                if self.current_symbol in self.unseen_networks:
                    self.unseen_networks.pop(self.current_symbol)
            values = line_parts[1]
        # No '=', so this is a value only line
        else:
            values = line_parts[0]  # values for previous var are continued this line
        for value_piece in values.split():
            if not value_piece:
                continue
            if not self.current_symbol:
                break
            if comment:
                self.unit.items.append(f"{value_piece} # {comment}")
            else:
                self.unit.items.append(value_piece)
                # token?
                if value_piece[0].isalpha() and ':' not in value_piece:
                    if definition_type == DEF_TYPE_SERVICES:
                        # already in top definitions list?
                        if value_piece not in self.services:
                            # already have it as an unused value?
                            if value_piece not in self.unseen_services:
                                self.unseen_services[value_piece] = True
                    if definition_type == DEF_TYPE_NETWORKS:
                        if value_piece not in self.networks:
                            if value_piece not in self.unseen_networks:
                                self.unseen_networks[value_piece] = True

    def ParseYaml(self, file: str, file_name: str) -> None:
        """Load network and service definitions from YAML.

        Arguments:
            file: A string containing the file contents.
            file_name: The original filename of the file.
        """

        try:
            file_data = yaml.load(file, Loader=SpanSafeYamlLoader(filename=file_name))
        except YAMLError as yaml_error:
            raise DefinitionFileTypeError(
                UserMessage("Unable to read file as YAML.", filename=file_name)
            ) from yaml_error

        self.ParseDefinitionsObject(file_data, file_name)

    def ParseDefinitionsObject(self, file_data: dict[str, Any], file_name: str) -> None:
        """Load network and service definitions from a Python object.

        Arguments:
            file_data: A Python dict where file_data.networks contains network
                data and/or file_data.services contains service data.
            file_name: The original filename of the file.
        """
        # Empty files are ignored with a warning
        if not file_data:
            logging.warning(UserMessage("Ignoring empty address book file.", filename=file_name))
            return

        # Check for at least one essential key, ignore with warning
        essential_keys = ['networks', 'services']

        if not any(key in file_data for key in essential_keys):
            logging.warning(
                UserMessage("File contains no network or service data.", filename=file_name)
            )
            return

        if 'networks' in file_data:
            self._ParseYamlNetworks(file_data, file_name)

        if 'services' in file_data:
            self._ParseYamlServices(file_data, file_name)

    def _ParseYamlNetworks(self, file_data: dict[str, Any], file_name: str) -> None:
        if 'networks' in file_data and not isinstance(file_data['networks'], dict):
            logging.warning(
                UserMessage(
                    "Network definition type error: dictionary expected.", filename=file_name
                )
            )
            return

        # Construct ItemUnit for each data point
        for symbol, symbol_def in file_data['networks'].items():
            if symbol in ["__line__", "__filename__"]:
                continue

            if not ('values' in symbol_def and isinstance(symbol_def['values'], list)):
                logging.info(
                    f'\nNetwork definition must be a list. Ignoring definition for network: {symbol}.'
                )
                continue

            if not self.token_re.match(symbol):
                logging.info(
                    f'\nNetwork name does not match recommended criteria: {symbol}\nOnly A-Z, a-z, 0-9, -, and _ allowed'
                )

            if symbol in self.networks:
                raise NamespaceCollisionError(
                    f'\nMultiple definitions found for network: {symbol}'
                )

            addr_unit = _ItemUnit(symbol)
            fqdn_unit = _ItemUnit(symbol)

            self.networks[symbol] = addr_unit
            self.fqdn[symbol] = fqdn_unit
            if symbol in self.unseen_networks:
                self.unseen_networks.pop(symbol)

            for item in symbol_def['values']:
                # 'item' can be:
                # 1. A string, understood as a network name reference
                # 2. A dictionary, with these fields:
                #    'address': A specific IP address or CIDR range
                #    'fqdn': A FQDN for use in DNS filtering.
                #    'name': A network name reference
                #    'comment': An optional comment
                # 'address' or 'name' must be present in any dictionary item
                value = None
                network_ref = None
                ip = None
                comment = None
                if isinstance(item, str):
                    value = network_ref = item
                elif isinstance(item, dict):
                    if 'name' in item and isinstance(item['name'], str):
                        value = network_ref = item['name']
                    elif 'address' in item and isinstance(item['address'], str):
                        value = ip = item['address']
                    elif 'fqdn' in item and isinstance(item['fqdn'], str):
                        value = item['fqdn']
                    else:
                        logging.info(f'\nNetwork name or CIDR expected for: {symbol}')
                        continue

                    if 'comment' in item and isinstance(item['comment'], str):
                        comment = item['comment']
                else:
                    logging.info(f'\nUnexpected symbol definition: {symbol}')
                    continue
                if comment is None:
                    if network_ref:
                        addr_unit.items.append(value)
                        fqdn_unit.items.append(value)
                    elif ip:
                        addr_unit.items.append(value)
                    else:
                        fqdn_unit.items.append(value)
                else:
                    if network_ref:
                        addr_unit.items.append(f'{value} # {comment}')
                        fqdn_unit.items.append(f'{value} # {comment}')
                    elif ip:
                        addr_unit.items.append(f'{value} # {comment}')
                    else:
                        fqdn_unit.items.append(f'{value} # {comment}')

                if network_ref and network_ref not in self.networks:
                    if network_ref not in self.unseen_networks:
                        self.unseen_networks[network_ref] = True

    def _ParseYamlServices(self, file_data: dict[str, Any], file_name: str) -> None:
        if 'services' in file_data and not isinstance(file_data['services'], dict):
            logging.warning(
                UserMessage(
                    "Service definition type error: dictionary expected.", filename=file_name
                )
            )
            return

        # Construct ItemUnit for each data point
        for symbol, symbol_def in file_data['services'].items():
            if symbol in ["__line__", "__filename__"]:
                continue

            if not isinstance(symbol_def, list):
                logging.info(
                    f'\nService definition must be a list. Ignoring definition for service: {symbol}.'
                )
                continue

            if not self.token_re.match(symbol):
                logging.info(
                    f'\nService name does not match recommended criteria: {symbol}\nOnly A-Z, a-z, 0-9, -, and _ allowed'
                )

            if symbol in self.services:
                raise NamespaceCollisionError(
                    f'\nMultiple definitions found for service: {symbol}'
                )

            unit = _ItemUnit(symbol)

            self.services[symbol] = unit
            if symbol in self.unseen_services:
                self.unseen_services.pop(symbol)

            for item in symbol_def:
                # 'item' can be:
                # 1. A string, understood as a service name reference
                # 2. A dictionary, with these fields:
                #    'protocol': A logical or numeric protocol (e.g. 'tcp')
                #    'port': A port number
                #    'name': A service name reference
                #    'comment': An optional comment
                # ('protocol' and 'port') or 'name' must be present in any dictionary item
                value = None
                service_ref = None
                service_port = None
                comment = None
                if isinstance(item, str):
                    value = service_ref = item
                elif isinstance(item, dict):
                    if 'name' in item and isinstance(item['name'], str):
                        value = service_ref = item['name']
                    elif (
                        'port' in item
                        and (isinstance(item['port'], str) or isinstance(item['port'], int))
                        and 'protocol' in item
                        and (
                            isinstance(item['protocol'], str) or isinstance(item['protocol'], int)
                        )
                    ):
                        protocol = item['protocol']
                        port = item['port']
                        value = service_port = f'{port}/{protocol}'
                    else:
                        logging.info(f'\nService name or port definition expected for: {symbol}')
                        continue
                    if 'comment' in item and isinstance(item['comment'], str):
                        comment = item['comment']
                else:
                    logging.info(f'\nUnexpected symbol definition: {symbol}')
                    continue

                if comment is None:
                    unit.items.append(value)
                else:
                    unit.items.append(f'{value} # {comment}')

                if service_ref and service_ref not in self.services:
                    if service_ref not in self.unseen_services:
                        self.unseen_services[service_ref] = True



# Copyright 2021 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Kubernetes NetworkPolicy resource generator.

More information about Kubernetes NetworkPolicy:
https://kubernetes.io/docs/concepts/services-networking/network-policies/
https://kubernetes.io/docs/tasks/administer-cluster/declare-network-policy/
"""

import copy
import re
from typing import TypedDict

import yaml
from absl import logging

from aerleon.lib import aclgenerator
from aerleon.lib.policy import Policy, Term


class Error(aclgenerator.Error):
    """Generic error class."""


class K8sNetworkPolicyError(Error):
    """Raised with problems in formatting for Kubernetes NetworkPolicies."""


class ExceededAttributeCountError(Error):
    """Raised when the total attribute count of a policy is above the maximum."""


class MatchExpression(TypedDict):
    key: str
    operator: str
    values: 'list[str]'


LabelSelector = TypedDict(
    "PodSelector", {"matchLabels": dict[str, str], "matchExpression": MatchExpression}
)


class IPBlock(TypedDict):
    cidr: str
    exceot: 'list[str]'


class PolicyPeer(TypedDict):
    podSelector: LabelSelector
    namespaceSelector: LabelSelector
    ipBlock: IPBlock


class PolicyPort(TypedDict):
    protocol: str
    port: int
    endPort: int


class Egress(TypedDict):
    ports: 'list[PolicyPort]'
    to: 'list[PolicyPeer]'


Ingress = TypedDict("Ingress", {"ports": "list[PolicyPort]", "from": "list[PolicyPeer]"})


class Spec(TypedDict):
    podSelector: LabelSelector
    policyTypes: 'list[str]'


class Annotations(TypedDict):
    comment: str
    owner: str


class Metadata(TypedDict):
    name: str
    annotations: Annotations


class NetworkPolicy(TypedDict):
    apiVersion: str
    kind: str
    metadata: Metadata
    spec: Spec


def IsDefaultDeny(term: Term) -> bool:
    """Returns true if a term is a default deny without IPs, ports, etc."""
    skip_attrs = [
        'flattened',
        'flattened_addr',
        'flattened_saddr',
        'flattened_daddr',
        'action',
        'comment',
        'name',
        'logging',
        'direction',
    ]
    if 'deny' not in term.action:
        return False
    # This lc will look through all methods and attributes of the object.
    # It returns only the attributes that need to be looked at to determine if
    # this is a default deny.
    for i in [
        a
        for a in dir(term)
        if not a.startswith('__') and a.islower() and not callable(getattr(term, a))
    ]:
        if i in skip_attrs:
            continue
        v = getattr(term, i)
        if isinstance(v, str) and v:
            return False
        if isinstance(v, list) and v:
            return False

    return True


def GetNextPriority(priority):
    """Get the priority for the next rule."""
    return priority


class Term(aclgenerator.Term):
    """Creates the term for the Kubernetes NetworkPolicy."""

    _API_VERSION = 'networking.k8s.io/v1'
    _RESOURCE_KIND = 'NetworkPolicy'
    # Policy rule name has to match specific RE:
    # No more than 253 characters, beginning and ending
    # with a lowercase alphanumeric character with dashes, dots, and lowercase
    # alphanumerics between.
    # Details: https://kubernetes.io/docs/concepts/overview/working-with-objects/names
    _TERM_NAME_RE = re.compile(r'^[a-z0-9]([a-z0-9-\.]){0,251}[a-z0-9]$')
    _TERM_MAX_LENGTH = 253

    # Protocols allowed are only tcp/udp/sctp
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#networkpolicyport-v1-networking-k8s-io
    PROTO_MAP = {
        'tcp': 6,
        'udp': 17,
        'sctp': 132,
        'all': -1,  # Used for default deny
    }

    def __init__(self, term: Term) -> None:
        super().__init__(term)
        self.term = term

        if 'deny' in self.term.action:
            if IsDefaultDeny(term):
                return
            else:
                raise K8sNetworkPolicyError(
                    'Kubernetes NetworkPolicy does not support explicit deny terms.'
                )

        self._validateDirection()

        if self.term.source_port:
            raise K8sNetworkPolicyError(
                'Kubernetes NetworkPolicy does not support source port restrictions.'
            )

        # Raise an error if the flattening of address exclusions would result in
        # overly broad network access control policies
        if (
            self.term.source_address_exclude
            and self.term.source_address
            or self.term.destination_address_exclude
            and self.term.destination_address
        ):
            self.term.FlattenAll(mutate=False)
            if not self.term.flattened_saddr and self.term.direction == 'INGRESS':
                logging.error(
                    'Kubernetes NetworkPolicy term %s no longer contains any source '
                    'addresses after the prefixes in source_address_exclude were '
                    'removed. Not rendering term.',
                    self.term.name,
                )
                self.term = None
                return

            if not self.term.flattened_daddr and self.term.direction == 'EGRESS':
                logging.error(
                    'Kubernetes NetworkPolicy term %s no longer contains any destination '
                    'addresses after the prefixes in destination_address_exclude were '
                    'removed. Not rendering term.',
                    self.term.name,
                )
                self.term = None
                return

    def __str__(self) -> str:
        """Convert term to a string."""
        return yaml.safe_dump(self.ConvertToDict())

    def _validateDirection(self) -> None:
        if self.term.direction == 'INGRESS':
            if not self.term.source_address:
                raise K8sNetworkPolicyError('Ingress rule missing required field "source-address"')

            if self.term.destination_address:
                raise K8sNetworkPolicyError(
                    'Ingress rules cannot include ' '"destination-address.'
                )

        elif self.term.direction == 'EGRESS':
            if self.term.source_address:
                raise K8sNetworkPolicyError('Egress rules cannot include "source-address".')

            if not self.term.destination_address:
                raise K8sNetworkPolicyError(
                    'Egress rule missing required field "destination-address".'
                )

    def ConvertToDict(
        self,
    ) -> NetworkPolicy:
        """Convert term to a dictionary.

        This is used to get a dictionary describing this term which can be
        output easily as a YAML object.

        Returns:
          A dictionary that contains a complete Kubernetes NetworkPolicy resource

        Raises:
          K8sNetworkPolicyError: The term name is not valid.
        """
        if not self.term:
            return {}

        if not self._TERM_NAME_RE.match(self.term.name):
            raise K8sNetworkPolicyError(
                'Term name %s is not valid. See https://kubernetes.io/docs/concepts/overview/working-with-objects/names for more information'
                % (self.term.name)
            )

        resource_dict = {
            'apiVersion': self._API_VERSION,
            'kind': self._RESOURCE_KIND,
            'metadata': {
                'name': self.term.name,
                'annotations': {},
            },
            'spec': {'podSelector': {}, 'policyTypes': [self.term.direction.capitalize()]},
        }

        if self.term.comment:
            resource_dict['metadata']['annotations']['comment'] = ' '.join(self.term.comment)
        if self.term.owner:
            resource_dict['metadata']['annotations']['owner'] = self.term.owner

        # We only allow one kind of deny policy, and thats a default deny. Because
        # of that, we can quickly return an empty policy in the specified direction
        if 'deny' in self.term.action:
            return resource_dict

        peer_selectors = []
        peer_selector_key = ''
        base_port_selector = {}
        if self.term.direction == 'INGRESS':
            for source_address in self.term.source_address:
                peer_selector = {'ipBlock': {'cidr': str(source_address)}}
                for exclude in self.term.source_address_exclude:
                    if peer_selector['ipBlock'].get('except') is None:
                        peer_selector['ipBlock']['except'] = []

                    peer_selector['ipBlock']['except'].append(str(exclude))
                peer_selectors.append(peer_selector)
            peer_selector_key = 'from'
        else:
            for destination_address in self.term.destination_address:
                peer_selector = {'ipBlock': {'cidr': str(destination_address)}}
                for exclude in self.term.destination_address_exclude:
                    if peer_selector['ipBlock'].get('except') is None:
                        peer_selector['ipBlock']['except'] = []

                    peer_selector['ipBlock']['except'].append(str(exclude))
                peer_selectors.append(peer_selector)
            peer_selector_key = 'to'

        # Build a base port selector list from ports
        base_port_selectors = []
        for start, end in self.term.destination_port:
            if start == end:
                base_port_selector = {'port': start}
            else:
                base_port_selector = {'port': start, 'endPort': end}

            base_port_selectors.append(base_port_selector)

        # Use the ports info to make one selector per port pair per proto
        port_selectors = []
        for proto in self.term.protocol:
            # If the list of ports is null, we still need to specify proto
            if not base_port_selectors:
                port_selectors.append({'protocol': proto.upper()})
                continue

            for base_selector in base_port_selectors:
                current_selector = copy.deepcopy(base_selector)
                # NetworkPolicies require uppercased proto name
                current_selector['protocol'] = proto.upper()
                port_selectors.append(current_selector)

        resource_dict['spec'][self.term.direction.lower()] = [
            {
                'ports': port_selectors,
                peer_selector_key: peer_selectors,
            }
        ]

        return resource_dict


class K8s(aclgenerator.ACLGenerator):
    """A Kubernetes NetworkPolicy object."""

    _API_VERSION = 'networking.k8s.io/v1'
    _RESOURCE_KIND = 'NetworkPolicyList'
    _PLATFORM = 'k8s'
    SUFFIX = '.yml'
    _SUPPORTED_AF = frozenset('mixed')
    _GOOD_DIRECTION = ['INGRESS', 'EGRESS']
    _OPTIONAL_SUPPORTED_KEYWORDS = frozenset(['expiration'])

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, _ = super()._BuildTokens()

        # add extra things
        supported_tokens |= {'expiration', 'owner'}

        # remove unsupported things
        supported_tokens -= {'icmp_type', 'source-port', 'verbatim'}
        # easier to make a new structure
        supported_sub_tokens = {'action': {'accept', 'deny'}}

        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: Policy, exp_info: int) -> None:
        self.network_policies = []
        total_rule_count = 0

        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)
            filter_name = header.FilterName(self._PLATFORM)

            direction = 'INGRESS'
            if filter_options:
                for i in self._GOOD_DIRECTION:
                    if i in filter_options:
                        direction = i
                        filter_options.remove(i)

            term_names = set()
            for term in terms:
                if term.stateless_reply:
                    logging.warning(
                        'WARNING: Term %s in policy %s is a stateless reply '
                        'term and will not be rendered.',
                        term.name,
                        filter_name,
                    )
                    continue
                if not term.comment:
                    term.comment = header.comment
                if direction == 'EGRESS':
                    term.name += '-e'
                term.name = self.FixTermLength(term.name)
                if term.name in term_names:
                    raise K8sNetworkPolicyError(f'Duplicate term name {term.name}')
                term_names.add(term.name)

                term.direction = direction
                if term.option:
                    raise K8sNetworkPolicyError(
                        'Kubernetes NetworkPolicy does not support term options.'
                    )

                term_dict = Term(term).ConvertToDict()
                if term_dict:
                    total_rule_count += 1
                    self.network_policies.append(term_dict)

        logging.info('Total rule count of policy %s is: %d', filter_name, total_rule_count)

    def __str__(self) -> str:
        if not self.network_policies:
            return ''
        list_resource = {
            'apiVersion': self._API_VERSION,
            'kind': self._RESOURCE_KIND,
            'items': self.network_policies,
        }
        return yaml.safe_dump(list_resource)
